package gosubmit

// type Option func(f *Filler) error

// func AddFile(fieldname string, filename string, contents []byte) Option {
// 	return func(f *Filler) error {
// 		f.AddFile(fieldname, filename, contents)
// 	}
// }
